# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest

from specfile.sections import Section
from specfile.tags import Comment, Comments, Tag, Tags


def test_find():
    tags = Tags(
        [
            Tag("Name", "test", "test", ": ", Comments()),
            Tag("Version", "0.1", "0.1", ": ", Comments()),
            Tag("Release", "1%{?dist}", "1.fc35", ": ", Comments()),
        ]
    )
    assert tags.find("version") == 1
    with pytest.raises(ValueError):
        tags.find("Epoch")


def test_parse():
    tags = Tags.parse(
        Section(
            "package",
            [
                "%global ver_major 1",
                "%global ver_minor 0",
                "",
                "# this is a test package",
                "# not to be used in production",
                "Name:    test",
                "Version: %{ver_major}.%{ver_minor}",
                "  # this is a valid comment",
                "Release: 1%{?dist}",
                "",
                "%if 0",
                "Epoch:   1",
                "%endif",
                "",
                "Requires:          make",
                "Requires(post):    bash",
            ],
        ),
        Section(
            "package",
            [
                "",
                "",
                "",
                "# this is a test package",
                "# not to be used in production",
                "Name:    test",
                "Version: 1.0",
                "  # this is a valid comment",
                "Release: 1.fc35",
                "",
                "",
                "",
                "",
                "",
                "Requires:          make",
                "Requires(post):    bash",
            ],
        ),
    )
    assert tags[0].name == "Name"
    assert tags[0].comments[0].text == "this is a test package"
    assert tags[0].comments[1].text == "not to be used in production"
    assert tags[1].name == "Version"
    assert tags[1].value == "%{ver_major}.%{ver_minor}"
    assert tags[1].valid
    assert tags[1].expanded_value == "1.0"
    assert not tags[1].comments
    assert tags.release.comments[0].prefix == "  # "
    assert tags.epoch.name == "Epoch"
    assert not tags.epoch.valid
    assert tags.requires.value == "make"
    assert "requires(post)" in tags
    assert tags[-1].name == "Requires(post)"


def test_get_raw_section_data():
    tags = Tags(
        [
            Tag(
                "Name",
                "test",
                "test",
                ":    ",
                Comments(
                    [
                        Comment("this is a test package"),
                        Comment("not to be used in production"),
                    ],
                    ["%global ver_major 1", "%global ver_minor 0", ""],
                ),
            ),
            Tag("Version", "%{ver_major}.%{ver_minor}", "1.0", ": ", Comments()),
            Tag(
                "Release",
                "1%{?dist}",
                "1.fc35",
                ": ",
                Comments([Comment("this is a valid comment", "  # ")]),
            ),
            Tag("Epoch", "1", "", ":   ", Comments([], ["", "%if 0"])),
            Tag(
                "Requires", "make", "make", ":          ", Comments([], ["%endif", ""])
            ),
            Tag("Requires(post)", "bash", "bash", ":    ", Comments()),
        ],
        [],
    )
    assert tags.get_raw_section_data() == [
        "%global ver_major 1",
        "%global ver_minor 0",
        "",
        "# this is a test package",
        "# not to be used in production",
        "Name:    test",
        "Version: %{ver_major}.%{ver_minor}",
        "  # this is a valid comment",
        "Release: 1%{?dist}",
        "",
        "%if 0",
        "Epoch:   1",
        "%endif",
        "",
        "Requires:          make",
        "Requires(post):    bash",
    ]


def test_copy_tags():
    tags = Tags(
        [
            Tag("Name", "test", "test", ": ", Comments()),
        ]
    )
    tags_copy = copy.deepcopy(tags)
    assert tags == tags_copy



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
from abc import ABC
from typing import Any, Dict, List, Optional, SupportsIndex, Union, cast, overload

from specfile.macro_options import MacroOptions
from specfile.sections import Section


class PrepMacro(ABC):
    """
    Class that represents a %prep macro.

    Attributes:
        name: Literal name of the macro.
        options: Options of the macro.
    """

    CANONICAL_NAME: str
    OPTSTRING: str
    DEFAULTS: Dict[str, Union[bool, int, str]]

    def __init__(
        self,
        name: str,
        options: MacroOptions,
        delimiter: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        preceding_lines: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `PrepMacro` object.

        Args:
            name: Literal name of the macro.
            options: Options of the macro.
            delimiter: Delimiter separating name and option string.
            prefix: Characters preceding the macro on a line.
            suffix: Characters following the macro on a line.
            preceding_lines: Lines of the %prep section preceding the macro.

        Returns:
            Constructed instance of `PrepMacro` class.
        """
        self.name = name
        self.options = options
        self._delimiter = delimiter
        self._prefix = prefix or ""
        self._suffix = suffix or ""
        self._preceding_lines = (
            preceding_lines.copy() if preceding_lines is not None else []
        )

    def __repr__(self) -> str:
        options = repr(self.options)
        preceding_lines = repr(self._preceding_lines)
        # determine class name dynamically so that inherited classes
        # don't have to reimplement __repr__()
        return (
            f"{self.__class__.__name__}('{self.name}', {options}, "
            f"'{self._delimiter}', '{self._prefix}', '{self._suffix}', "
            f"{preceding_lines})"
        )

    def get_raw_data(self) -> List[str]:
        options = str(self.options)
        # ensure delimiter is not empty when there are any options
        if options and not self._delimiter:
            self._delimiter = " "
        return self._preceding_lines + [
            f"{self._prefix}{self.name}{self._delimiter}{options}{self._suffix}"
        ]


class SetupMacro(PrepMacro):
    """Class that represents a %setup macro."""

    CANONICAL_NAME: str = "%setup"
    OPTSTRING: str = "a:b:cDn:Tq"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {
        "n": "%{name}-%{version}",
    }


class PatchMacro(PrepMacro):
    """Class that represents a %patch macro."""

    CANONICAL_NAME: str = "%patch"
    OPTSTRING: str = "P:p:REb:z:F:d:o:Z"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {}

    @property
    def number(self) -> int:
        """Number of the %patch macro."""
        if self.options.P is not None:
            return int(self.options.P)
        tokens = re.split(r"(\d+)", self.name, maxsplit=1)
        if len(tokens) > 1:
            return int(tokens[1])
        return 0

    @number.setter
    def number(self, value: int) -> None:
        if self.options.P is not None:
            self.options.P = value
            return
        self.name = f"{self.CANONICAL_NAME}{value}"


class AutosetupMacro(PrepMacro):
    """Class that represents an %autosetup macro."""

    CANONICAL_NAME: str = "%autosetup"
    OPTSTRING: str = "a:b:cDn:TvNS:p:"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {
        "n": "%{name}-%{version}",
        "S": "patch",
    }


class AutopatchMacro(PrepMacro):
    """Class that represents an %autopatch macro."""

    CANONICAL_NAME: str = "%autopatch"
    OPTSTRING: str = "vp:m:M:"
    DEFAULTS: Dict[str, Union[bool, int, str]] = {}


class PrepMacros(collections.UserList):
    """
    Class that represents a list of %prep macros.

    Attributes:
        data: List of individual %prep macros.
    """

    def __init__(
        self,
        data: Optional[List[PrepMacro]] = None,
        remainder: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `PrepMacros` object.

        Args:
            data: List of individual %prep macros.
            remainder: Leftover lines in the section.

        Returns:
            Constructed instance of `PrepMacros` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._remainder = remainder.copy() if remainder is not None else []

    def __repr__(self) -> str:
        data = repr(self.data)
        remainder = repr(self._remainder)
        return f"PrepMacros({data}, {remainder})"

    def __contains__(self, item: object) -> bool:
        if isinstance(item, type):
            return any(isinstance(m, item) for m in self.data)
        return any(
            m.name.startswith(item) if item == "%patch" else m.name == item
            for m in self.data
        )

    @overload
    def __getitem__(self, i: SupportsIndex) -> PrepMacro:
        pass

    @overload
    def __getitem__(self, i: slice) -> "PrepMacros":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return PrepMacros(self.data[i], self._remainder)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            preceding_lines = self.data[index]._preceding_lines.copy()
            del self.data[index]
            if index < len(self.data):
                self.data[index]._preceding_lines = (
                    preceding_lines + self.data[index]._preceding_lines
                )
            else:
                self._remainder = preceding_lines + self._remainder

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def __getattr__(self, name: str) -> PrepMacro:
        if not self.valid_prep_macro(name):
            return super().__getattribute__(name)
        try:
            return self.data[self.find(f"%{name}")]
        except ValueError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        if not self.valid_prep_macro(name):
            return super().__delattr__(name)
        try:
            self.__delitem__(self.find(f"%{name}"))
        except ValueError:
            raise AttributeError(name)

    @staticmethod
    def valid_prep_macro(name: str) -> bool:
        return name in ("setup", "autosetup", "autopatch") or name.startswith("patch")

    def copy(self) -> "PrepMacros":
        return PrepMacros(self.data, self._remainder)

    def find(self, name: str) -> int:
        for i, macro in enumerate(self.data):
            if macro.name == name:
                return i
        raise ValueError

    def get_raw_data(self) -> List[str]:
        result = []
        for macro in self.data:
            result.extend(macro.get_raw_data())
        result.extend(self._remainder)
        return result


class Prep(collections.abc.Container):
    """
    Class that represents a %prep section.

    Attributes:
        macros: List of individual %prep macros.
    """

    def __init__(self, macros: PrepMacros) -> None:
        self.macros = macros.copy()

    def __repr__(self) -> str:
        macros = repr(self.macros)
        return f"Prep({macros})"

    def __contains__(self, item: object) -> bool:
        return item in self.macros

    def __getattr__(self, name: str) -> PrepMacro:
        if not self.macros.valid_prep_macro(name):
            return super().__getattribute__(name)
        return getattr(self.macros, name)

    def __delattr__(self, name: str) -> None:
        if not self.macros.valid_prep_macro(name):
            return super().__delattr__(name)
        return delattr(self.macros, name)

    def add_patch_macro(self, number: int, **kwargs: Any) -> None:
        """
        Adds a new %patch macro with given number and options.

        Args:
            number: Macro number.
            P: The -P option (patch number).
            p: The -p option (strip number).
            R: The -R option (reverse).
            E: The -E option (remove empty files).
            b: The -b option (backup).
            z: The -z option (suffix).
            F: The -F option (fuzz factor).
            d: The -d option (working directory).
            o: The -o option (output file).
            Z: The -Z option (set UTC times).
        """
        options = MacroOptions([], PatchMacro.OPTSTRING, PatchMacro.DEFAULTS)
        for k, v in kwargs.items():
            setattr(options, k, v)
        macro = PatchMacro(PatchMacro.CANONICAL_NAME, options, " ")
        macro.number = number
        index, _ = min(
            ((i, m) for i, m in enumerate(self.macros) if isinstance(m, PatchMacro)),
            key=lambda im: abs(im[1].number - number),
            default=(len(self.macros), None),
        )
        if (
            index < len(self.macros)
            and cast(PatchMacro, self.macros[index]).number <= number
        ):
            index += 1
        self.macros.insert(index, macro)

    def remove_patch_macro(self, number: int) -> None:
        """
        Removes a %patch macro with given number.

        Args:
            number: Macro number.
        """
        index = next(
            (
                i
                for i, m in enumerate(self.macros)
                if isinstance(m, PatchMacro) and m.number == number
            ),
            None,
        )
        if index:
            del self.macros[index]

    @classmethod
    def parse(cls, section: Section) -> "Prep":
        """
        Parses a section into a `Prep` object.

        Args:
            section: %prep section.

        Returns:
            Constructed instance of `Prep` class.
        """
        # match also macros enclosed in conditionalized macro expansion
        # e.g.: %{?with_system_nss:%patch30 -p3 -b .nss_pkcs11_v3}
        macro_regex = re.compile(
            r"(?P<c>%{!?\?\w+:)?.*?"
            r"(?P<m>%(setup|patch\d*|autopatch|autosetup))"
            r"(?P<d>\s*)"
            r"(?P<o>.*?)"
            r"(?(c)}|$)"
        )
        data = []
        buffer: List[str] = []
        for line in section:
            m = macro_regex.search(line)
            if m:
                name, delimiter, option_string = (
                    m.group("m"),
                    m.group("d"),
                    m.group("o"),
                )
                prefix, suffix = line[: m.start("m")], line[m.end("o") :]
                klass = next(
                    (
                        klass
                        for klass in PrepMacro.__subclasses__()
                        if name.startswith(klass.CANONICAL_NAME)
                    ),
                    None,
                )
                if not klass:
                    buffer.append(line)
                    continue
                options = MacroOptions(
                    MacroOptions.tokenize(option_string),
                    klass.OPTSTRING,
                    klass.DEFAULTS,
                )
                data.append(klass(name, options, delimiter, prefix, suffix, buffer))
                buffer = []
            else:
                buffer.append(line)
        return cls(PrepMacros(data, buffer))

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from `Prep` object.

        Returns:
            List of lines forming the reconstructed section data.
        """
        return self.macros.get_raw_data()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
from typing import List, Optional, SupportsIndex, Union, cast, overload

from specfile.constants import SECTION_NAMES

# name for the implicit "preamble" section
PREAMBLE = "package"


class Section(collections.UserList):
    """
    Class that represents a spec file section.

    Attributes:
        name: Name of the section (without the leading '%').
        data: List of lines forming the content of the section, not including newline characters.
    """

    def __init__(self, name: str, data: Optional[List[str]] = None) -> None:
        super().__init__()
        if not name or name.split()[0].lower() not in SECTION_NAMES:
            raise ValueError(f"Invalid section name: '{name}'")
        self.name = name
        if data is not None:
            self.data = data.copy()

    def __str__(self) -> str:
        data = "".join(f"{i}\n" for i in self.data)
        if self.name == PREAMBLE:
            return data
        return f"%{self.name}\n{data}"

    def __repr__(self) -> str:
        data = repr(self.data)
        return f"Section('{self.name}', {data})"

    def __copy__(self) -> "Section":
        return Section(self.name, self.data)

    @overload
    def __getitem__(self, i: SupportsIndex) -> str:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Section":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Section(self.name, self.data[i])
        else:
            return self.data[i]

    @property
    def normalized_name(self) -> str:
        """Normalized name of the section. All characters are lowercased."""
        return self.name.lower()

    def copy(self) -> "Section":
        return Section(self.name, self.data)

    def get_raw_data(self) -> List[str]:
        if self.name == PREAMBLE:
            return self.data
        return [f"%{self.name}"] + self.data


class Sections(collections.UserList):
    """
    Class that represents all spec file sections, hence the entire spec file.

    Sections can be accessed by index or conveniently by name as attributes:
    ```
    # print the third line of the first section
    print(sections[0][2])

    # remove the last line of %prep section
    del sections.prep[-1]

    # replace the entire %prep section
    sections.prep = ['line 1', 'line 2']

    # delete %changelog
    del sections.changelog
    ```

    Attributes:
        data: List of individual sections. Preamble is expected to always be the first.
    """

    def __str__(self) -> str:
        return "".join(str(i) for i in self.data)

    def __repr__(self) -> str:
        data = repr(self.data)
        return f"Sections({data})"

    def __copy__(self) -> "Sections":
        return Sections(self.data)

    def __contains__(self, name: object) -> bool:
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            data = super().__getattribute__("data")
        except AttributeError:
            return False
        return any(s.name.lower() == cast(str, name).split()[0].lower() for s in data)

    def __getattr__(self, name: str) -> Section:
        if name not in self:
            return super().__getattribute__(name)
        try:
            return self.get(name)
        except ValueError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Union[Section, List[str]]) -> None:
        if name not in self:
            return super().__setattr__(name, value)
        try:
            if isinstance(value, Section):
                self.data[self.find(name)] = value
            else:
                self.data[self.find(name)].data = value
        except ValueError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        if name not in self:
            return super().__delattr__(name)
        try:
            del self.data[self.find(name)]
        except ValueError:
            raise AttributeError(name)

    def get(self, name: str) -> Section:
        return self.data[self.find(name)]

    def find(self, name: str) -> int:
        name = name.lower()
        for i, section in enumerate(self.data):
            if section.name.lower() == name:
                return i
        raise ValueError

    @classmethod
    def parse(cls, lines: List[str]) -> "Sections":
        """
        Parses given lines into sections.

        Args:
            lines: Lines to parse.

        Returns:
            Constructed instance of `Sections` class.
        """
        section_name_regexes = [
            re.compile(rf"^%{re.escape(n)}(\s+.*$|$)", re.IGNORECASE)
            for n in SECTION_NAMES
        ]
        section_starts = []
        for i, line in enumerate(lines):
            if line.startswith("%"):
                for r in section_name_regexes:
                    if r.match(line):
                        section_starts.append(i)
                        break
        section_starts.append(len(lines))
        data = [Section(PREAMBLE, lines[: section_starts[0]])]
        for start, end in zip(section_starts, section_starts[1:]):
            data.append(Section(lines[start][1:], lines[start + 1 : end]))
        return cls(data)

    def get_raw_data(self) -> List[str]:
        result = []
        for section in self.data:
            result.extend(section.get_raw_data())
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import datetime
import re
from typing import List, Optional, SupportsIndex, Union, overload

import rpm

from specfile.exceptions import SpecfileException
from specfile.sections import Section
from specfile.utils import EVR

WEEKDAYS = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
MONTHS = (
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
)


class ChangelogEntry:
    """
    Class that represents a changelog entry.

    Attributes:
        header: Header of the entry.
        content: List of lines forming the content of the entry.
    """

    def __init__(
        self,
        header: str,
        content: List[str],
        following_lines: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `ChangelogEntry` object.

        Args:
            header: Header of the entry.
            content: List of lines forming the content of the entry.
            following_lines: Extra lines that follow the entry.

        Returns:
            Constructed instance of `ChangelogEntry` class.
        """
        self.header = header
        self.content = content.copy()
        self._following_lines = (
            following_lines.copy() if following_lines is not None else []
        )

    def __str__(self) -> str:
        return f"{self.header}\n" + "\n".join(self.content) + "\n"

    def __repr__(self) -> str:
        content = repr(self.content)
        following_lines = repr(self._following_lines)
        return f"ChangelogEntry('{self.header}', {content}, {following_lines})"

    @property
    def evr(self) -> Optional[str]:
        """EVR (Epoch, Version, Release) of the entry."""
        m = re.match(
            r"""
            ^.*
            \s+                       # preceding whitespace
            ((?P<sb>\[)|(?P<rb>\())?  # optional opening bracket
            (?P<evr>(\d+:)?\S+-\S+?)  # EVR
            (?(sb)\]|(?(rb)\)))       # matching closing bracket
            :?                        # optional colon
            \s*                       # optional following whitespace
            $
            """,
            self.header,
            re.VERBOSE,
        )
        if not m:
            return None
        return m.group("evr")

    @property
    def extended_timestamp(self) -> bool:
        """Whether the timestamp present in the entry header is extended (date and time)."""
        weekdays = "|".join(WEEKDAYS)
        months = "|".join(MONTHS)
        m = re.search(
            rf"""
            ({weekdays})     # weekday
            [ ]
            ({months})       # month
            [ ]+
            ([12]?\d|3[01])  # day of month
            [ ]
            ([01]\d|2[0-3])  # hour
            :
            ([0-5]\d)        # minute
            :
            ([0-5]\d)        # second
            [ ]
            \S+              # timezone
            [ ]
            \d{{4}}          # year
            """,
            self.header,
            re.VERBOSE,
        )
        return m is not None

    @property
    def day_of_month_padding(self) -> str:
        """Padding of day of month in the entry header timestamp"""
        weekdays = "|".join(WEEKDAYS)
        months = "|".join(MONTHS)
        m = re.search(
            rf"""
            ({weekdays})                 # weekday
            [ ]
            ({months})                   # month
            [ ]
            (?P<wsp>[ ]*)                # optional whitespace padding
            ((?P<zp>0)?\d|[12]\d|3[01])  # possibly zero-padded day of month
            """,
            self.header,
            re.VERBOSE,
        )
        if not m:
            return ""
        return m.group("wsp") + (m.group("zp") or "")

    @classmethod
    def assemble(
        cls,
        timestamp: Union[datetime.date, datetime.datetime],
        author: str,
        content: List[str],
        evr: Optional[str] = None,
        day_of_month_padding: str = "0",
        append_newline: bool = True,
    ) -> "ChangelogEntry":
        """
        Assembles a changelog entry.

        Args:
            timestamp: Timestamp of the entry.
              Supply `datetime` rather than `date` for extended format.
            author: Author of the entry.
            content: List of lines forming the content of the entry.
            evr: EVR (epoch, version, release) of the entry.
            day_of_month_padding: Padding to apply to day of month in the timestamp.
            append_newline: Whether the entry should be followed by an empty line.

        Returns:
            Constructed instance of `ChangelogEntry` class.
        """
        weekday = WEEKDAYS[timestamp.weekday()]
        month = MONTHS[timestamp.month - 1]
        header = f"* {weekday} {month}"
        if day_of_month_padding.endswith("0"):
            header += f" {day_of_month_padding[:-1]}{timestamp.day:02}"
        else:
            header += f" {day_of_month_padding}{timestamp.day}"
        if isinstance(timestamp, datetime.datetime):
            # extended format
            if not timestamp.tzinfo:
                timestamp = timestamp.replace(tzinfo=datetime.timezone.utc)
            header += f" {timestamp:%H:%M:%S %Z}"
        header += f" {timestamp:%Y} {author}"
        if evr is not None:
            header += f" - {evr}"
        return cls(header, content, [""] if append_newline else None)


class Changelog(collections.UserList):
    """
    Class that represents a changelog.

    Attributes:
        data: List of individual entries.
    """

    def __init__(
        self,
        data: Optional[List[ChangelogEntry]] = None,
        predecessor: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `Changelog` object.

        Args:
            data: List of individual changelog entries.
            predecessor: Lines at the beginning of a section that can't be parsed
              into changelog entries.

        Returns:
            Constructed instance of `Changelog` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._predecessor = predecessor.copy() if predecessor is not None else []

    def __str__(self) -> str:
        return "\n".join(str(i) for i in reversed(self.data))

    def __repr__(self) -> str:
        data = repr(self.data)
        predecessor = repr(self._predecessor)
        return f"Changelog({data}, {predecessor})"

    @overload
    def __getitem__(self, i: SupportsIndex) -> ChangelogEntry:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Changelog":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Changelog(self.data[i], self._predecessor)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            following_lines = self.data[index]._following_lines.copy()
            del self.data[index]
            if index < len(self.data):
                self.data[index]._following_lines.extend(following_lines[1:])
            else:
                self._predecessor.extend(following_lines[1:])

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def filter(
        self, since: Optional[str] = None, until: Optional[str] = None
    ) -> "Changelog":
        """
        Filters changelog entries with EVR between since and until.

        Args:
            since: Optional lower bound. If specified, entries with EVR higher
              than or equal to this will be included.
            until: Optional upper bound. If specified, entries with EVR lower
              than or equal to this will be included.

        Returns:
            Filtered changelog.
        """

        def parse_evr(s):
            try:
                evr = EVR.from_string(s)
            except SpecfileException:
                return "0", "0", ""
            return str(evr.epoch), evr.version or "0", evr.release

        if since is None:
            start_index = 0
        else:
            start_index = next(
                (
                    i
                    for i, e in enumerate(self.data)
                    if rpm.labelCompare(parse_evr(e.evr), parse_evr(since)) >= 0
                ),
                len(self.data) + 1,
            )
        if until is None:
            end_index = len(self.data) + 1
        else:
            end_index = next(
                (
                    i + 1
                    for i, e in reversed(list(enumerate(self.data)))
                    if rpm.labelCompare(parse_evr(e.evr), parse_evr(until)) <= 0
                ),
                0,
            )
        return self[start_index:end_index]

    @classmethod
    def parse(cls, section: Section) -> "Changelog":
        """
        Parses a %changelog section.

        Args:
            section: Section to parse.

        Returns:
            Constructed instance of `Changelog` class.
        """

        def extract_following_lines(content):
            following_lines: List[str] = []
            while content and not content[-1].strip():
                following_lines.insert(0, content.pop())
            return following_lines

        data: List[ChangelogEntry] = []
        predecessor = []
        header = None
        content: List[str] = []
        for line in section:
            if line.startswith("*"):
                if header:
                    following_lines = extract_following_lines(content)
                    data.insert(0, ChangelogEntry(header, content, following_lines))
                header = line
                content = []
            elif header:
                content.append(line)
            else:
                predecessor.append(line)
        if header:
            following_lines = extract_following_lines(content)
            data.insert(0, ChangelogEntry(header, content, following_lines))
        return cls(data, predecessor)

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from changelog.

        Returns:
            List of lines forming the reconstructed section data.
        """
        result = self._predecessor
        for entry in reversed(self.data):
            result.append(entry.header)
            result.extend(entry.content)
            result.extend(entry._following_lines)
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from specfile.exceptions import SpecfileException
from specfile.sourcelist import Sourcelist, SourcelistEntry
from specfile.sources import ListSource, Patches, Sources, TagSource
from specfile.tags import Comments, Tag, Tags


@pytest.mark.parametrize(
    "tag_name, number",
    [
        ("Source", None),
        ("Source0", "0"),
        ("Patch0001", "0001"),
        ("Patch28", "28"),
        ("Patch99999", "99999"),
    ],
)
def test_tag_source_extract_number(tag_name, number):
    ts = TagSource(Tag(tag_name, "", "", "", Comments()))
    assert ts._extract_number() == number


@pytest.mark.parametrize(
    "tags, default, result",
    [
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source2", "source2"),
            ],
            True,
            False,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
                ("Source", "source1"),
                ("Source", "source2"),
            ],
            False,
            True,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
            ],
            False,
            False,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source", "source0"),
            ],
            True,
            True,
        ),
    ],
)
def test_sources_detect_implicit_numbering(tags, default, result):
    sources = Sources(
        Tags([Tag(t, v, v, ": ", Comments()) for t, v in tags]),
        [],
        default_to_implicit_numbering=default,
    )
    assert sources._detect_implicit_numbering() == result


@pytest.mark.parametrize(
    "ref_name, ref_separator, number, name, separator",
    [
        ("Source", ": ", 28, "Source", ": "),
        ("Source0", ": ", 28, "Source28", ":"),
        ("Source0001", ":      ", 2, "Source0002", ":      "),
    ],
)
def test_sources_get_tag_format(ref_name, ref_separator, number, name, separator):
    sources = Sources(Tags(), [])
    reference = TagSource(Tag(ref_name, "", "", ref_separator, Comments()))
    assert sources._get_tag_format(reference, number) == (name, separator)


@pytest.mark.parametrize(
    "tags, number, index",
    [
        (["Name", "Version"], 0, 2),
        ([], 999, 0),
    ],
)
def test_sources_get_initial_tag_setup(tags, number, index):
    sources = Sources(
        Tags([Tag(t, "test", "test", ": ", Comments()) for t in tags]), []
    )
    assert sources._get_initial_tag_setup(number) == (index, f"Source{number}", ": ")


@pytest.mark.parametrize(
    "tags, deduplicated_tags",
    [
        (["Source", "Source"], ["Source", "Source"]),
        (["Source0", "Source"], ["Source0", "Source"]),
        (
            ["Source0", "Source0", "Source", "Source"],
            ["Source0", "Source1", "Source", "Source"],
        ),
        (
            ["Source100", "Source100", "Source", "Source1028", "Source1011"],
            ["Source100", "Source101", "Source", "Source1028", "Source1011"],
        ),
        (
            ["Source2", "Source2", "Source4", "Source", "Source999"],
            ["Source2", "Source3", "Source4", "Source", "Source999"],
        ),
        (
            ["Source3", "Source4", "Source", "Source5"],
            ["Source3", "Source4", "Source", "Source6"],
        ),
        (
            [
                "Source",
                "Source",
                "Source100",
                "Source101",
                "Source101",
                "Source102",
                "Source",
            ],
            [
                "Source",
                "Source",
                "Source100",
                "Source101",
                "Source102",
                "Source103",
                "Source",
            ],
        ),
    ],
)
def test_sources_deduplicate_tag_names(tags, deduplicated_tags):
    sources = Sources(
        Tags([Tag(t, "test", "test", ": ", Comments()) for t in tags]), []
    )
    sources._deduplicate_tag_names()
    assert [t.name for t in sources._tags] == deduplicated_tags


@pytest.mark.parametrize(
    "tags, sourcelists, index, location, number, cls",
    [
        ([("Name", "test"), ("Version", "0.1")], [], 0, "test", 0, TagSource),
        ([("Name", "test"), ("Version", "0.1")], [[]], 0, "test", 0, ListSource),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            10,
            "test",
            2,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            10,
            "test",
            4,
            ListSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            1,
            "test",
            1,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source10", "source10"),
            ],
            [["source11", "source12"]],
            1,
            "test",
            10,
            TagSource,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            2,
            "test",
            2,
            ListSource,
        ),
        (
            [("Name", "test"), ("Version", "0.1"), ("Source0", "source0")],
            [],
            10,
            "source0",
            None,
            None,
        ),
    ],
)
def test_sources_insert(tags, sourcelists, index, location, number, cls):
    sources = Sources(
        Tags([Tag(t, v, v, ": ", Comments()) for t, v in tags]),
        [
            Sourcelist([SourcelistEntry(s, Comments()) for s in sl])
            for sl in sourcelists
        ],
    )
    if location in [v for t, v in tags if t.startswith(Sources.prefix)] + [
        s for sl in sourcelists for s in sl
    ]:
        with pytest.raises(SpecfileException):
            sources.insert(index, location)
    else:
        sources.insert(index, location)
        if index >= len(sources):
            index = len(sources) - 1
        assert isinstance(sources[index], cls)
        assert sources[index].number == number
        assert sources[index].location == location


@pytest.mark.parametrize(
    "tags, number, location, index",
    [
        ([("Name", "test"), ("Version", "0.1")], 28, "test", 0),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            0,
            "test",
            0,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            2,
            "test",
            2,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            42,
            "test",
            3,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
                ("Source28", "source28"),
                ("Source999", "source999"),
            ],
            1000,
            "test",
            4,
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source28", "source28"),
                ("Source42", "source42"),
                ("Source5", "source5"),
                ("Source", "source"),
            ],
            37,
            "test",
            1,
        ),
    ],
)
def test_sources_insert_numbered(tags, number, location, index):
    sources = Sources(Tags([Tag(t, v, v, ": ", Comments()) for t, v in tags]), [])
    if location in [v for t, v in tags if t.startswith(Sources.prefix)]:
        with pytest.raises(SpecfileException):
            sources.insert_numbered(number, location)
    else:
        assert sources.insert_numbered(number, location) == index
        assert isinstance(sources[index], TagSource)
        assert sources[index].number == number
        assert sources[index].location == location


@pytest.mark.parametrize(
    "tags, sourcelists, number, new_tags, new_sourcelists",
    [
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            1,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
            ],
            [],
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
            2,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [],
        ),
        (
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source2", "source3"]],
            2,
            [
                ("Name", "test"),
                ("Version", "0.1"),
                ("Source0", "source0"),
                ("Source1", "source1"),
            ],
            [["source3"]],
        ),
    ],
)
def test_sources_remove_numbered(tags, sourcelists, number, new_tags, new_sourcelists):
    tags = Tags([Tag(t, v, v, ": ", Comments()) for t, v in tags])
    sourcelists = [
        Sourcelist([SourcelistEntry(s, Comments()) for s in sl]) for sl in sourcelists
    ]
    sources = Sources(tags, sourcelists)
    sources.remove_numbered(number)
    assert tags == Tags([Tag(t, v, v, ": ", Comments()) for t, v in new_tags])
    assert sourcelists == [
        Sourcelist([SourcelistEntry(s, Comments()) for s in sl])
        for sl in new_sourcelists
    ]


@pytest.mark.parametrize(
    "ref_name, ref_separator, number, name, separator",
    [
        ("Patch99", ":      ", 100, "Patch100", ":     "),
        ("Patch9999", ":  ", 28, "Patch28", ":    "),
        ("Patch0999", ":  ", 28, "Patch0028", ":  "),
        ("Source2", ":     ", 0, "Patch0", ":      "),
    ],
)
def test_patches_get_tag_format(ref_name, ref_separator, number, name, separator):
    patches = Patches(Tags(), [])
    reference = TagSource(Tag(ref_name, "", "", ref_separator, Comments()))
    assert patches._get_tag_format(reference, number) == (name, separator)


@pytest.mark.parametrize(
    "tags, number, index",
    [
        (["Name", "Version", "Source0"], 0, 3),
        (["Name", "Version", "Source0", "BuildRequires"], 1, 3),
        (["Name", "Version"], 2, 2),
        ([], 999, 0),
    ],
)
def test_patches_get_initial_tag_setup(tags, number, index):
    patches = Patches(
        Tags([Tag(t, "test", "test", ": ", Comments()) for t in tags]), []
    )
    flexmock(patches).should_receive("_get_tag_format").and_return(
        f"Patch{number}", ": "
    )
    assert patches._get_initial_tag_setup(number) == (index, f"Patch{number}", ": ")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
import string
from enum import Enum, auto
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union, overload

from specfile.exceptions import MacroOptionsException


class TokenType(Enum):
    DEFAULT = auto()
    WHITESPACE = auto()
    QUOTED = auto()
    DOUBLE_QUOTED = auto()


class Token(collections.abc.Hashable):
    """
    Class that represents a single token in an option string.

    Attributes:
        type: Token type.
        value: Token value.
    """

    def __init__(self, type: TokenType, value: str) -> None:
        self.type = type
        self.value = value

    def __repr__(self) -> str:
        type = repr(self.type)
        return f"Token({type}, '{self.value}')"

    def __str__(self) -> str:
        if self.type == TokenType.WHITESPACE:
            return self.value
        elif self.type == TokenType.QUOTED:
            # escape single quotes
            value = self.value.replace("'", r"\'")
            return f"'{value}'"
        elif self.type == TokenType.DOUBLE_QUOTED:
            # escape double quotes
            value = self.value.replace('"', r"\"")
            return f'"{value}"'
        # escape quotes and whitespace
        return re.sub(r"['\"\s]", r"\\\g<0>", self.value)

    def _key(self) -> tuple:
        return self.type, self.value

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Token):
            return NotImplemented
        return self._key() == other._key()


class Positionals(collections.abc.MutableSequence):
    """Class that represents a sequence of positional arguments."""

    def __init__(self, options: "MacroOptions") -> None:
        """
        Constructs a `Positionals` object.

        Args:
            options: MacroOptions instance this object is tied with.

        Returns:
            Constructed instance of `Positionals` class.
        """
        self._options = options

    def __repr__(self) -> str:
        options = repr(self._options)
        return f"Positionals({options})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (Positionals, list)):
            return NotImplemented
        return list(self) == other

    def __len__(self) -> int:
        return len(self._get_items())

    @overload
    def __getitem__(self, i: int) -> Union[int, str]:
        pass

    @overload
    def __getitem__(self, i: slice) -> List[Union[int, str]]:
        pass

    def __getitem__(self, i):
        items = self._get_items()
        if isinstance(i, slice):
            result = []
            for index in items[i]:
                value = self._options._tokens[index].value
                result.append(int(value) if value.isnumeric() else value)
            return result
        else:
            value = self._options._tokens[items[i]].value
            return int(value) if value.isnumeric() else value

    @overload
    def __setitem__(self, i: int, item: Union[int, str]) -> None:
        pass

    @overload
    def __setitem__(self, i: slice, item: Iterable[Union[int, str]]) -> None:
        pass

    def __setitem__(self, i, item):
        items = self._get_items()
        if isinstance(i, slice):
            for i0, i1 in enumerate(range(len(items))[i]):
                self._options._tokens[items[i1]].value = str(item[i0])
        else:
            self._options._tokens[items[i]].value = str(item)

    def __delitem__(self, i: Union[int, slice]) -> None:
        def delete(index):
            tokens = self._options._tokens
            if index == 0:
                if len(tokens) > 1 and tokens[1].type == TokenType.WHITESPACE:
                    del tokens[1]
            else:
                if tokens[index - 1].type == TokenType.WHITESPACE:
                    index -= 1
                    del tokens[index]
            del tokens[index]

        items = self._get_items()
        if isinstance(i, slice):
            for index in reversed(items[i]):
                delete(index)
        else:
            delete(items[i])

    def _get_items(self) -> List[int]:
        """
        Gets all positional arguments.

        Returns:
            List of indices of tokens that are positional arguments.
        """
        result = []
        i = 0
        while i < len(self._options._tokens):
            if self._options._tokens[i].type == TokenType.WHITESPACE:
                i += 1
                continue
            # exclude options (starting with -) and their arguments (if any)
            value = self._options._tokens[i].value
            if value.startswith("-"):
                i += 1
                if len(value) > 1:
                    if value[1] in self._options.optstring.replace(":", ""):
                        if self._options._requires_argument(value[1]):
                            if len(value) == 2:
                                if (
                                    i < len(self._options._tokens)
                                    and self._options._tokens[i].type
                                    == TokenType.WHITESPACE
                                ):
                                    i += 1
                                i += 1
                continue
            result.append(i)
            i += 1
        return result

    def insert(self, i: int, value: Union[int, str]) -> None:
        """
        Inserts a new positional argument at a specified index.

        Args:
            i: Requested index.
            value: Value of the positional argument.
        """
        items = self._get_items()
        if i > len(items):
            i = len(items)
        if items and i < len(items):
            index = items[i]
            if index > 0:
                if self._options._tokens[index - 1].type == TokenType.WHITESPACE:
                    index -= 1
        else:
            index = len(self._options._tokens)
        self._options._tokens.insert(
            index,
            Token(
                TokenType.DOUBLE_QUOTED
                if self._options._needs_quoting(value)
                else TokenType.DEFAULT,
                str(value),
            ),
        )
        if index > 0:
            self._options._tokens.insert(index, Token(TokenType.WHITESPACE, " "))


class MacroOptions(collections.abc.MutableMapping):
    """
    Class that represents macro options.

    Attributes:
        optstring: getopt-like option string containing recognized option characters.
          Option characters are ASCII letters, upper or lower-case.
          If such a character is followed by a colon, the option
          requires an argument.
        defaults: Dict specifying default arguments to options.
    """

    def __init__(
        self,
        tokens: List[Token],
        optstring: Optional[str] = None,
        defaults: Optional[Dict[str, Union[bool, int, str]]] = None,
    ) -> None:
        """
        Constructs a `MacroOptions` object.

        Args:
            tokens: List of tokens in an option string.
            optstring: String containing recognized option characters.
              Option characters are ASCII letters, upper or lower-case.
              If such a character is followed by a colon, the option
              requires an argument.
            defaults: Dict specifying default arguments to options.

        Returns:
            Constructed instance of `MacroOptions` class.
        """
        self._tokens = tokens.copy()
        self.optstring = optstring or ""
        self.defaults = defaults.copy() if defaults is not None else {}

    def __repr__(self) -> str:
        tokens = repr(self._tokens)
        defaults = repr(self.defaults)
        return f"MacroOptions({tokens}, '{self.optstring}', {defaults})"

    def __str__(self) -> str:
        return "".join(str(t) for t in self._tokens)

    def _valid_option(self, name: str) -> bool:
        """
        Determines if a name represents a recognized option.

        Args:
            name: Name of the option.

        Returns:
            True if the option is recognized, otherwise False.
        """
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            optstring = super().__getattribute__("optstring")
        except AttributeError:
            return False
        return name in optstring.replace(":", "")

    def _requires_argument(self, option: str) -> bool:
        """
        Determines if an option requires an argument.

        Args:
            option: Name of the option.

        Returns:
            True if the option requires an argument, otherwise False.

        Raises:
            ValueError if the specified option is not valid.
        """
        i = self.optstring.index(option) + 1
        return i < len(self.optstring) and self.optstring[i] == ":"

    def _find_option(self, name: str) -> Tuple[Optional[int], Optional[int]]:
        """
        Searches for an option in tokens of an option string.

        Args:
            name: Name of the option.

        Returns:
            Tuple of indices where the first is the index of a token matching
            the option and the second is the index of a token matching
            its argument, or None if there is no match.
        """
        option = f"-{name}"
        for i, token in reversed(list(enumerate(self._tokens))):
            if not token.value.startswith(option):
                continue
            if token.value != option:
                return i, i
            if not self._requires_argument(name):
                return i, None
            j = i + 1
            if j == len(self._tokens):
                return i, None
            if self._tokens[j].type == TokenType.WHITESPACE:
                j += 1
            if j == len(self._tokens):
                return i, None
            if self._tokens[j].value.startswith("-"):
                return i, None
            return i, j
        return None, None

    @staticmethod
    def _needs_quoting(value):
        # if there is a whitespace, enquote the value rather than escaping it
        return any(ws in str(value) for ws in string.whitespace)

    def __getattr__(self, name: str) -> Optional[Union[bool, int, str]]:
        if not self._valid_option(name):
            return super().__getattribute__(name)
        i, j = self._find_option(name)
        if i is None:
            if self._requires_argument(name):
                return self.defaults.get(name)
            return False
        value = (
            self._tokens[j].value
            if j is not None and j > i
            else self._tokens[i].value[2:]
        )
        if not value:
            return True
        if value.isnumeric():
            return int(value)
        return value

    def __setattr__(self, name: str, value: Union[bool, int, str]) -> None:
        if not self._valid_option(name):
            return super().__setattr__(name, value)
        if self._requires_argument(name) and isinstance(value, bool):
            raise MacroOptionsException(f"Option -{name} requires an argument.")
        if (
            not self._requires_argument(name)
            and not isinstance(value, bool)
            and value is not None
        ):
            raise MacroOptionsException(f"Option -{name} is a flag.")
        i, j = self._find_option(name)
        if i is None:
            if value is not None and value is not False:
                if self._tokens:
                    self._tokens.append(Token(TokenType.WHITESPACE, " "))
                if value is True:
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}"))
                elif isinstance(value, int):
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}{value}"))
                else:
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}"))
                    self._tokens.append(Token(TokenType.WHITESPACE, " "))
                    self._tokens.append(
                        Token(
                            TokenType.DOUBLE_QUOTED
                            if self._needs_quoting(value)
                            else TokenType.DEFAULT,
                            value,
                        )
                    )
            return
        if value is None or value is False:
            return delattr(self, name)
        if j is not None and j > i:
            if self._needs_quoting(value):
                if self._tokens[j].type not in (
                    TokenType.QUOTED,
                    TokenType.DOUBLE_QUOTED,
                ):
                    self._tokens[j].type = TokenType.DOUBLE_QUOTED
            else:
                self._tokens[j].type = TokenType.DEFAULT
            self._tokens[j].value = str(value)
        else:
            if self._needs_quoting(value):
                self._tokens[i].value = self._tokens[i].value[:2]
                self._tokens.insert(i + 1, Token(TokenType.DOUBLE_QUOTED, str(value)))
                self._tokens.insert(i + 1, Token(TokenType.WHITESPACE, " "))
            else:
                self._tokens[i].value = self._tokens[i].value[:2] + str(value)

    def __delattr__(self, name: str) -> None:
        if not self._valid_option(name):
            return super().__delattr__(name)
        i, j = self._find_option(name)
        if i is None:
            return
        if i == 0:
            if (
                j is not None
                and j + 1 < len(self._tokens)
                and self._tokens[j + 1].type == TokenType.WHITESPACE
            ):
                j += 1
        elif self._tokens[i - 1].type == TokenType.WHITESPACE:
            i -= 1
        if j is not None and j > i:
            del self._tokens[i : j + 1]
        else:
            del self._tokens[i]

    def __len__(self) -> int:
        return len(
            {
                t.value[1]
                for t in self._tokens
                if t.type != TokenType.WHITESPACE
                and t.value.startswith("-")
                and len(t.value) > 1
                and t.value[1] in self.optstring.replace(":", "")
            }
        )

    def __getitem__(self, key: str) -> Union[bool, int, str]:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key: str, item: Union[bool, int, str]) -> None:
        try:
            return setattr(self, key, item)
        except AttributeError:
            raise KeyError(key)

    def __delitem__(self, key: str) -> None:
        try:
            return delattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        for option in self.optstring.replace(":", ""):
            i, _ = self._find_option(option)
            if i is not None:
                yield option

    @property
    def positional(self) -> Positionals:
        """Sequence of positional arguments."""
        return Positionals(self)

    @positional.setter
    def positional(self, value: List[Union[int, str]]) -> None:
        positionals = Positionals(self)
        positionals.clear()
        positionals.extend(value)

    @staticmethod
    def tokenize(option_string: str) -> List[Token]:
        """
        Tokenizes an option string.

        Follows the same rules as poptParseArgvString() that is used by RPM.

        Args:
            option_string: Option string.

        Returns:
            List of tokens.

        Raises:
            MacroOptionsException if the option string is untokenizable.
        """
        result = []
        token = ""
        quote = None
        inp = list(option_string)
        while inp:
            c = inp.pop(0)
            if c == quote:
                if token:
                    result.append(
                        Token(
                            TokenType.QUOTED
                            if quote == "'"
                            else TokenType.DOUBLE_QUOTED,
                            token,
                        )
                    )
                    token = ""
                quote = None
                continue
            if quote:
                if c == "\\":
                    if not inp:
                        raise MacroOptionsException("No escaped character")
                    c = inp.pop(0)
                    if c != quote:
                        token += "\\"
                token += c
                continue
            if c.isspace():
                if token:
                    result.append(Token(TokenType.DEFAULT, token))
                    token = ""
                whitespace = c
                while inp:
                    c = inp.pop(0)
                    if not c.isspace():
                        break
                    whitespace += c
                inp.insert(0, c)
                result.append(Token(TokenType.WHITESPACE, whitespace))
                continue
            if c in ('"', "'"):
                if token:
                    result.append(Token(TokenType.DEFAULT, token))
                    token = ""
                quote = c
                continue
            if c == "\\":
                if not inp:
                    raise MacroOptionsException("No escaped character")
                c = inp.pop(0)
            token += c
        if quote:
            raise MacroOptionsException("No closing quotation")
        if token:
            result.append(Token(TokenType.DEFAULT, token))
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest

from specfile.sections import Section, Sections


def test_find():
    sections = Sections(
        [
            Section("package"),
            Section("prep"),
            Section("changelog"),
            Section("pAckage foo"),
        ]
    )
    assert sections.find("prep") == 1
    assert sections.find("package foo") == 3
    with pytest.raises(ValueError):
        sections.find("install")


def test_get():
    sections = Sections([Section("package"), Section("package bar")])
    assert sections.get("package") == []
    assert sections.get("package bar") == []
    with pytest.raises(ValueError):
        sections.get("package foo")


def test_parse():
    sections = Sections.parse(
        [
            "0",
            "",
            "",
            "%prep",
            "0",
            "1",
            "2",
            "",
            "%package x",
            "%files y",
            "0",
            "%changelog",
        ]
    )
    assert sections[0][0] == "0"
    assert sections[1].name == "prep"
    assert sections.prep == ["0", "1", "2", ""]
    assert sections[2].name == "package x"
    assert not sections[2]
    assert sections[-1].name == "changelog"


def test_parse_case_insensitive():
    sections = Sections.parse(
        ["0", "", "%Prep", "0", "1", "2", "", "%pAckage x", "Requires: bar"]
    )
    assert sections[0][0] == "0"
    assert sections[1].name == "Prep"
    assert sections.prep == ["0", "1", "2", ""]
    assert sections[2].name == "pAckage x"
    assert sections[2] == ["Requires: bar"]


def test_parse_invalid_name():
    sections = Sections.parse(
        [
            "%description",
            "This is a description.",
            "",
            "%description(fr)",
            "Ceci est une description.",
            "",
        ]
    )
    assert len(sections) == 2  # including empty preamble
    assert sections[1].name == "description"
    assert sections.description[2] == "%description(fr)"


def test_copy_sections():
    sections = Sections([Section("package"), Section("package bar")])
    sections_copy = copy.deepcopy(sections)
    assert sections == sections_copy
